searchNodes=[{"type":"module","title":"MuonTrap","doc":"MuonTrap protects you from lost and out of control OS processes. You can use it as a System.cmd/3 replacement or to pull OS processes into an Erlang supervision tree via MuonTrap.Daemon . Either way, if the Erlang process that runs the command dies, then the OS processes will die as well. MuonTrap tries very hard to kill OS processes so that remnants don't hang around the system when your Erlang code thinks they should be gone. MuonTrap can use the Linux kernel's cgroup feature to contain the child process and all of its children. From there, you can limit CPU and memory and other resources to the process group. MuonTrap does not require cgroups but keep in mind that OS processes can escape. It is, however, still an improvement over System.cmd/3 which does not have a mechanism for dealing it OS processes that do not monitor their stdin for when to close. For more information, see the documentation for MuonTrap.cmd/3 and MuonTrap.Daemon Configuring cgroups On most Linux distributions, use cgcreate to create a new cgroup. You can name them almost anything. The command below creates one named muontrap for the current user. It supports memory and CPU controls. sudo cgcreate -a $(whoami) -g memory,cpu:muontrap Nerves systems do not contain cgcreate by default. Due to the simpler Linux setup, it may be sufficient to run File.mkdir_p(cgroup_path) to create a cgroup. For example: File . mkdir_p ( &quot;/sys/fs/cgroup/memory/muontrap&quot; ) This creates the cgroup path, muontrap under the memory controller. If you do not have the &quot;/sys/fs/cgroup&quot; directory, you will need to mount it or update your erlinit.config to mount it for you. See a newer official system for an example.","ref":"MuonTrap.html"},{"type":"function","title":"MuonTrap.cmd/3","doc":"Executes a command like System.cmd/3 via the muontrap wrapper. Options :cgroup_controllers - run the command under the specified cgroup controllers. Defaults to [] . :cgroup_base - create a temporary path under the specified cgroup path :cgroup_path - explicitly specify a path to use. Use :cgroup_base , unless you must control the path. :cgroup_sets - set a cgroup controller parameter before running the command :delay_to_sigkill - milliseconds before sending a SIGKILL to a child process if it doesn't exit with a SIGTERM (default 500 ms) :uid - run the command using the specified uid or username :gid - run the command using the specified gid or group The following System.cmd/3 options are also available: :into - injects the result into the given collectable, defaults to &quot;&quot; :cd - the directory to run the command in :env - an enumerable of tuples containing environment key-value as binary :arg0 - sets the command arg0 :stderr_to_stdout - redirects stderr to stdout when true :parallelism - when true , the VM will schedule port tasks to improve parallelism in the system. If set to false , the VM will try to perform commands immediately, improving latency at the expense of parallelism. The default can be set on system startup by passing the &quot;+spp&quot; argument to --erl . Examples Run a command: iex&gt; MuonTrap . cmd ( &quot;echo&quot; , [ &quot;hello&quot; ] ) { &quot;hello \\n &quot; , 0 } The next examples only run on Linux. To try this out, create new cgroups: sudo cgcreate -a $(whoami) -g memory,cpu:muontrap Run a command, but limit memory so severely that it doesn't work (for demo purposes, obviously): iex - donttest &gt; MuonTrap . cmd ( &quot;echo&quot; , [ &quot;hello&quot; ] , cgroup_controllers : [ &quot;memory&quot; ] , cgroup_path : &quot;muontrap/test&quot; , cgroup_sets : [ { &quot;memory&quot; , &quot;memory.limit_in_bytes&quot; , &quot;8192&quot; } ] ) { &quot;&quot; , 1 }","ref":"MuonTrap.html#cmd/3"},{"type":"function","title":"MuonTrap.muontrap_path/0","doc":"Return the absolute path to the muontrap executable. Call this if you want to invoke the muontrap port binary manually.","ref":"MuonTrap.html#muontrap_path/0"},{"type":"module","title":"MuonTrap.Daemon","doc":"Wrap an OS process in a GenServer so that it can be supervised. For example, in your children list add MuonTrap.Daemon like this: children = [ { MuonTrap.Daemon , [ &quot;my_server&quot; , [ &quot;--options&quot; , &quot;foo&quot; ) ] , [ cd : &quot;/some_directory&quot; ] ] } ] opts = [ strategy : :one_for_one , name : MyApplication.Supervisor ] Supervisor . start_link ( children , opts ) In the child_spec tuple, the second element is a list that corresponds to the MuonTrap.cmd/3 parameters. I.e., The first item in the list is the program to run, the second is a list of commandline arguments, and the third is a list of options. The same options as MuonTrap.cmd/3 are available with the following additions: :name - Name the Daemon GenServer :log_output - When set, send output from the command to the Logger. Specify the log level (e.g., :debug ) :log_prefix - Prefix each log message with this string (defaults to the program's path) :stderr_to_stdout - When set to true , redirect stderr to stdout. Defaults to false . :exit_status_to_reason - Optional function to convert the exit status (a number) to stop reason for the Daemon GenServer. Use if error exit codes carry information or aren't errors. If you want to run multiple MuonTrap.Daemon s under one supervisor, they'll all need unique IDs. Use Supervisor.child_spec/2 like this: Supervisor . child_spec ( { MuonTrap.Daemon , [ &quot;my_server&quot; , [ ] ] } , id : :server1 )","ref":"MuonTrap.Daemon.html"},{"type":"function","title":"MuonTrap.Daemon.cgget/3","doc":"Get the value of the specified cgroup variable.","ref":"MuonTrap.Daemon.html#cgget/3"},{"type":"function","title":"MuonTrap.Daemon.cgset/4","doc":"Modify a cgroup variable.","ref":"MuonTrap.Daemon.html#cgset/4"},{"type":"function","title":"MuonTrap.Daemon.child_spec/1","doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"MuonTrap.Daemon.html#child_spec/1"},{"type":"function","title":"MuonTrap.Daemon.os_pid/1","doc":"Return the OS pid to the muontrap executable.","ref":"MuonTrap.Daemon.html#os_pid/1"},{"type":"function","title":"MuonTrap.Daemon.start_link/3","doc":"Start/link a deamon GenServer for the specified command.","ref":"MuonTrap.Daemon.html#start_link/3"},{"type":"function","title":"MuonTrap.Daemon.statistics/1","doc":"Return statistics about the daemon Statistics: :output_byte_count - bytes output by the process being run","ref":"MuonTrap.Daemon.html#statistics/1"},{"type":"module","title":"MuonTrap.Options","doc":"Validate and normalize the options passed to MuonTrap.cmd/3 and MuonTrap.Daemon.start_link/3 This module is generally not called directly, but it's likely the source of exceptions if any options aren't quite right. Call validate/4 directly to debug or check options without invoking a command.","ref":"MuonTrap.Options.html"},{"type":"function","title":"MuonTrap.Options.validate/4","doc":"Validate options and normalize them for invoking commands Pass in :cmd or :daemon for the first parameter to allow function-specific options.","ref":"MuonTrap.Options.html#validate/4"},{"type":"type","title":"MuonTrap.Options.t/0","doc":"The following fields are always present: :cmd - the command to run :args - a list of arguments to the command The next fields are optional: :into - MuonTrap.cmd/3 only :cd :arg0 :stderr_to_stdout :parallelism :env :name - MuonTrap.Daemon -only :log_output - MuonTrap.Daemon -only :log_prefix - MuonTrap.Daemon -only :log_transform - MuonTrap.Daemon -only :stdio_window :exit_status_to_reason - MuonTrap.Daemon -only :cgroup_controllers :cgroup_path :cgroup_base :delay_to_sigkill :cgroup_sets :uid :gid","ref":"MuonTrap.Options.html#t:t/0"},{"type":"extras","title":"MuonTrap","doc":"Keep programs, deamons, and applications launched from Erlang and Elixir contained and well-behaved. This lightweight library kills OS processes if the Elixir process running them crashes and if you're running on Linux, it can use cgroups to prevent many other shenanigans. Some other features: Attach your OS process to a supervision tree via a convenient child_spec Set cgroup controls like thresholds on memory and CPU utilization Start OS processes as a different user or group Send SIGKILL to processes that aren't responsive to SIGTERM With cgroups , ensure that all children of launched processes have been killed too","ref":"readme.html"},{"type":"extras","title":"MuonTrap - TL;DR","doc":"Add muontrap to your project's mix.exs dependency list: def deps do [ { :muontrap , &quot;~&gt; 1.0&quot; } ] end Run a command similar to System.cmd/3 : iex&gt; MuonTrap . cmd ( &quot;echo&quot; , [ &quot;hello&quot; ] ) { &quot;hello \\n &quot; , 0 } Attach a long running process to a supervision tree using a child_spec like the following: { MuonTrap.Daemon , [ &quot;long_running_command&quot; , [ &quot;arg1&quot; , &quot;arg2&quot; ] , options ] } Running on Linux and can use cgroups? Then create a new cgroup: sudo cgcreate -a $(whoami) -g memory:mycgroup { MuonTrap.Daemon , [ &quot;long_running_command&quot; , [ &quot;arg1&quot; , &quot;arg2&quot; ] , [ cgroup_controllers : [ &quot;memory&quot; ] , cgroup_base : &quot;mycgroup&quot; ] ] } MuonTrap will create a cgroup under &quot;mycgroup&quot; to run the &quot;long_running_command&quot; . If the command fails, it will be restarted. If it should no longer be running (like if something else crashed in Elixir and supervision needs to clean up) then MuonTrap will kill &quot;long_running_command&quot; and all of its children. Want to know more? Read on...","ref":"readme.html#tl-dr"},{"type":"extras","title":"MuonTrap - The problem","doc":"The Erlang VM's port interface lets Elixir applications run external programs. This is important since it's not practical to rewrite everything in Elixir. Plus, if the program is long running like a daemon or a server, you use Elixir to supervise it and restart it on crashes. The catch is that the Erlang VM expects port processes to be well-behaved. As you'd expect, many useful programs don't quite meet the Erlang VM's expectations. For example, let's say that you want to monitor a network connection and decide that ping is the right tool. Here's how you could start ping in a process. iex&gt; pid = spawn ( fn -&gt; System . cmd ( &quot;ping&quot; , [ &quot;-i&quot; , &quot;5&quot; , &quot;localhost&quot; ] , into : IO . stream ( :stdio , :line ) ) end ) # PID &lt; 0 . 6116 . 0 &gt; PING localhost ( 127.0 . 0.1 ) : 56 data bytes 64 bytes from 127.0 . 0.1 : icmp_seq = 0 ttl = 64 time = 0.032 ms 64 bytes from 127.0 . 0.1 : icmp_seq = 1 ttl = 64 time = 0.077 ms To see that ping is running, call ps to look for it. You can also do this from a separate terminal window outside of IEx: iex&gt; :os . cmd ( &#39;ps -ef | grep ping&#39; ) |&gt; IO . puts 501 38820 38587 0 9 : 26 PM ?? 0 : 00.01 / sbin / ping - i 5 localhost 501 38824 38822 0 9 : 27 PM ?? 0 : 00.00 grep ping :ok Now exit the Elixir process. Imagine here that in the real program that something happened in Elixir and the process needs to exit and be restarted by a supervisor. iex&gt; Process . exit ( pid , :oops ) true iex&gt; :os . cmd ( &#39;ps -ef | grep ping&#39; ) |&gt; IO . puts 501 38820 38587 0 9 : 26 PM ?? 0 : 00.02 / sbin / ping - i 5 localhost 501 38833 38831 0 9 : 34 PM ?? 0 : 00.00 grep ping As you can tell, ping is still running after the exit. If you run :observer you'll see that Elixir did indeed terminate both the process and the port, but that didn't stop ping . The reason for this is that ping doesn't pay attention to stdin and doesn't notice the Erlang VM closing it to signal that it should exit. Imagine now that the process was supervised and it restarts. If this happens a regularly, you could be running dozens of ping commands. This is just one of the problems that muontrap fixes.","ref":"readme.html#the-problem"},{"type":"extras","title":"MuonTrap - Applicability","doc":"This is intended for long running processes. It's not great for interactive programs that communicate via the port or send signals. That feature is possible to add, but you'll probably be happier with other solutions like erlexec .","ref":"readme.html#applicability"},{"type":"extras","title":"MuonTrap - Running commands","doc":"The simplest way to use muontrap is as a replacement to System.cmd/3 . Here's an example using ping : iex&gt; pid = spawn ( fn -&gt; MuonTrap . cmd ( &quot;ping&quot; , [ &quot;-i&quot; , &quot;5&quot; , &quot;localhost&quot; ] , into : IO . stream ( :stdio , :line ) ) end ) # PID &lt; 0 . 30860 . 0 &gt; PING localhost ( 127.0 . 0.1 ) : 56 data bytes 64 bytes from 127.0 . 0.1 : icmp_seq = 0 ttl = 64 time = 0.027 ms 64 bytes from 127.0 . 0.1 : icmp_seq = 1 ttl = 64 time = 0.081 ms Now if you exit that process, ping gets killed as well: iex&gt; Process . exit ( pid , :oops ) true iex&gt; :os . cmd ( &#39;ps -ef | grep ping&#39; ) |&gt; IO . puts 501 38898 38896 0 9 : 58 PM ?? 0 : 00.00 grep ping :ok","ref":"readme.html#running-commands"},{"type":"extras","title":"MuonTrap - Containment with cgroups","doc":"Even if you don't make use of any cgroup controller features, having your port process contained can be useful just to make sure that everything is cleaned up on exit including any subprocesses. To set this up, first create a cgroup with appropriate permissions. Any path will do; muontrap just needs to be able to create a subdirectory underneath it for its use. For example: sudo cgcreate -a $(whoami) -g memory,cpu:mycgroup Be sure to create the group for all of the cgroup controllers that you wish to use with muontrap . The above example creates it for the memory and cpu controllers. In Elixir, call MuonTrap.cmd/3 with the cgroup options now. In this case, we'll use the cpu controller, but this example would work fine with any of the controllers. iex&gt; MuonTrap . cmd ( &quot;spawning_program&quot; , [ ] , cgroup_controllers : [ &quot;cpu&quot; ] , cgroup_base : &quot;mycgroup&quot; ) { &quot;hello \\n &quot; , 0 } In this example, muontrap runs spawning_program in a sub-cgroup under the cpu/mycgroup group. The cgroup parameters may be modified outside of muontrap using cgset or my accessing the cgroup mountpoint manually. On any error or if the Erlang VM closes the port or if spawning_program exits, muontrap will kill all OS processes in cgroup. No need to worry about random processes accumulating on your system. Note that if you use cgroup_base , a temporary cgroup is created for running the command. If you want muontrap to use a particular cgroup and not create a subgroup for the command, use the :cgroup_path option. Note that if you explicitly specify a cgroup, be careful not to use it for anything else. MuonTrap assumes that it owns the cgroup and when it needs to kill processes, it kills all of them in the cgroup. Limit the memory used by a process Linux's cgroups are very powerful and the examples here only scratch the surface. If you'd like to limit an OS process and all of its child processes to a maximum amount of memory, you can do that with the memory controller: iex&gt; MuonTrap . cmd ( &quot;memory_hog&quot; , [ ] , cgroup_controllers : [ &quot;memory&quot; ] , cgroup_base : &quot;mycgroup&quot; , cgroup_sets : [ { &quot;memory&quot; , &quot;memory.limit_in_bytes&quot; , &quot;268435456&quot; } ] ) That line restricts the total memory used by memory_hog to 256 MB. Limit CPU usage in a port Limiting the maximum CPU usage is also possible. Two parameters control that with the cpu controller: cpu.cfs_period_us specifies the number of microseconds in the scheduling period and cpu.cfs_quota_us specifies how many of those microseconds can be used. Here's an example call that prevents a program from using more than 50% of the CPU: iex&gt; MuonTrap . cmd ( &quot;cpu_hog&quot; , [ ] , cgroup_controllers : [ &quot;cpu&quot; ] , cgroup_base : &quot;mycgroup&quot; , cgroup_sets : [ { &quot;cpu&quot; , &quot;cpu.cfs_period_us&quot; , &quot;100000&quot; } , { &quot;cpu&quot; , &quot;cpu.cfs_quota_us&quot; , 50000 } ] )","ref":"readme.html#containment-with-cgroups"},{"type":"extras","title":"MuonTrap - Supervision","doc":"For many long running programs, you may want to restart them if they crash. Luckily Erlang already has mechanisms to do this. MuonTrap provides a GenServer called MuonTrap.Daemon that you can hook into one of your supervision trees. For example, you could specify it like this in your application's supervisor: def start ( _type , _args ) do children = [ { MuonTrap.Daemon , [ &quot;command&quot; , [ &quot;arg1&quot; , &quot;arg2&quot; ] , options ] } ] opts = [ strategy : :one_for_one , name : MyApp.Supervisor ] Supervisor . start_link ( children , opts ) end Supervisors provide three restart strategies, :permanent , :temporary , and :transient . They work as follows: :permanent - Always restart the command if it exits or crashes. Restarts are limited to the Supervisor's restart intensity settings as they would be with normal GenServer s. This is the default. :transient - If the exit status of the command is 0 (i.e., success), then don't restart. Any other exit status is considered an error and the command is restarted. :temporary - Don't restart If you're running more than one MuonTrap.Daemon under the same Supervisor , then you'll need to give each one a unique :id . Here's an example child_spec for setting the :id and the :restart parameters: Supervisor . child_spec ( { MuonTrap.Daemon , [ &quot;command&quot; , [ &quot;arg1&quot; ] , options ] } , id : :my_daemon , restart : :transient )","ref":"readme.html#supervision"},{"type":"extras","title":"MuonTrap - stdio flow control","doc":"The Erlang port feature does not implement flow control from messages coming from the port process. Since MuonTrap captures stdio from the program being run, it's possible that the program sends output so fast that it grows the Elixir process's mailbox big enough to cause an out-of-memory error. MuonTrap protects against this by implementing a flow control mechanism. When triggered, the running program's stdout and stderr file handles won't be read and hence it will eventually be blocked from writing to those handles. The :stdio_window option specifies the maximum number of unacknowledged bytes allowed. The default is 10 KB.","ref":"readme.html#stdio-flow-control"},{"type":"extras","title":"MuonTrap - muontrap development","doc":"In order to run the tests, some additional tools need to be installed. Specifically the cgcreate and cgget binaries need to be installed (and available on $PATH ). Typically the package may be called cgroup-tools (on arch linux you need to install the libcgroup aur package). Then run: sudo cgcreate -a $(whoami) -g memory,cpu:muontrap_test","ref":"readme.html#muontrap-development"},{"type":"extras","title":"MuonTrap - License","doc":"All original source code in this project is licensed under Apache-2.0. Additionally, this project follows the REUSE recommendations and labels so that licensing and copyright are clear at the file level. Exceptions to Apache-2.0 licensing are: Configuration and data files are licensed under CC0-1.0 Documentation is CC-BY-4.0","ref":"readme.html#license"}]